import logging
from time import perf_counter

import numpy as np
from coyote import add_task


class Timer:
    def __enter__(self):
        self.start = perf_counter()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.elapsed_time = perf_counter() - self.start


class DataHandler:
    def __init__(self, var, t0, cell_slice):
        self.var = var
        self.cell_slice = cell_slice
        self.loco_server = None
        mi, ma, st = cell_slice.indices(var.shape[-1])
        assert st == 1, "non-contiguous cell slices are not supported"
        self.t = t0
        self.prune_offset = var.attrs.get("loco::num_retained_timesteps", None)
        self.t_flushed = self.t
        self.time_chunk_size = var.chunks[0]
        self.use_buffer = var.attrs.get("hiopy::use_buffer", self.time_chunk_size > 1)
        if self.use_buffer:
            self.buf = np.empty([self.time_chunk_size, *var.shape[1:-1], ma - mi], dtype=var.dtype)

    def _create_flush_task(self):
        fro = self.t_flushed % self.time_chunk_size
        to = fro + (self.t - self.t_flushed)
        buf = self.buf[fro:to, ...].copy()
        t_flushed = int(self.t_flushed)
        t = int(self.t)

        def _flush():
            nonlocal t, t_flushed, buf, self
            if self.prune_offset is not None:
                prune_from = max(0, t_flushed - self.prune_offset)
                prune_to = max(0, t - self.prune_offset)
                with Timer() as timer:
                    self.var[prune_from:prune_to, ..., self.cell_slice] = self.var.fill_value
                logging.info(
                    f"pruned {self.var.name} for timesteps {fro}:{to}. Took {timer.elapsed_time}s"
                )

            with Timer() as timer:
                self.var[t_flushed:t, ..., self.cell_slice] = buf
            logging.info(
                f"wrote {self.var.name} for timesteps {t_flushed}:{t}."
                f" Took {timer.elapsed_time}s"
            )
            if self.loco_server is not None:
                self.loco_server.set_timestep(self.var.name, t - 1)

        self.t_flushed = self.t
        return _flush

    def flush(self):
        self._create_flush_task()()

    def __call__(self, event):
        logging.info(
            f"received {self.var.name} at {self.t % self.time_chunk_size+1}/{self.time_chunk_size}"
        )
        if self.use_buffer:
            self.buf[self.t % self.time_chunk_size, ...] = np.transpose(event.data)
            event.release()
        else:
            self.buf = np.squeeze(np.transpose(event.data))[None, ...]
        self.t += 1
        if self.t % self.time_chunk_size == 0 or not self.use_buffer:  # chunk complete
            add_task(self._create_flush_task())

    def set_loco_server(self, loco_server):
        self.loco_server = loco_server
