#!/usr/bin/env python3

import logging
from math import ceil, prod
import numpy as np

logger = logging.getLogger(__name__)


def _map_weighted(a, b, wa=None, wb=None):
    """
    Maps elements of a to b based on their weighted contributions.

    Parameters:
    - a: List of first elements to map
    - b: List of second elements to map
    - wa: Weights for elements in a (default is uniform weight)
    - wb: Weights for elements in b (default is uniform weight)

    Returns:
    A dictionary mapping each element in a to the corresponding elements in b.
    """
    wa = np.ones(len(a)) if wa is None else wa
    wb = np.ones(len(b)) if wb is None else wb

    # Calculate cumulative weights for a and b
    partial_sum_a = np.cumsum(wa)
    partial_sum_b = np.cumsum(wb)

    factor = partial_sum_a[-1] / partial_sum_b[-1]  # Normalization factor
    indices = [0, *np.searchsorted(factor * partial_sum_b, partial_sum_a, side="right")]

    return {a[ai]: b[i:j] for ai, (i, j) in enumerate(zip(indices[:-1], indices[1:]))}


def _estimate_size(*data_vars):
    """
    Estimates the size in bytes of one data chunk of given data variables.

    Parameters:
    - data_vars: List of data variable objects with shape and chunk size attributes.

    Returns:
    Total estimated size in bytes of the data variables.

    """
    size = 0.0
    for dv in data_vars:
        if dv.time is not None:
            dt = dv.time[1] - dv.time[0]
            size += prod(dv.shape[1:]) * dv.dtype.itemsize / dt
    return size


def _distribute_chunks(var, size):
    """
    Distributes chunks of a data variable across ranks.

    Parameters:
    - var: Data variable to distribute chunks from.
    - size: Number of ranks to distribute to.

    Returns:
    A list of slice objects representing distributed chunks for each rank.
    """
    if size == 1:
        return [slice(None)]  # Return all data for a single rank

    # Get chunk size and total number of cells
    cell_chunk_size = var.chunks[-1]
    ncells = var.shape[-1]
    nchunks = ceil(ncells / cell_chunk_size)

    assert nchunks // size > 0, "Cannot distribute chunks (too many ranks)"

    # Map ranks to chunks
    ranks_to_chunks = _map_weighted(range(int(size)), range(nchunks))
    return [
        slice(cell_chunk_size * c[0], min(ncells, cell_chunk_size * (c[-1] + 1)))
        for i, c in sorted(ranks_to_chunks.items())
    ]


def distribute_work(grouped_data_vars, size):
    """
    Distributes work among ranks based on grouped data variables.

    Parameters:
    - grouped_data_vars: A dictionary grouping group_names (grid) to data variable objects.
    - size: Number of ranks to distribute work to.

    Returns:
    A list of tuples, each representing the distribution of work to a specific rank.
    """
    if size == 0:
        assert len(grouped_data_vars) == 0, "Cannot distribute non-empty dataset on 0 processes"
        return []

    # Estimate sizes for each group of variables
    group_sizes = {g: _estimate_size(*v) for g, v in grouped_data_vars.items()}
    total_size = sum(group_sizes.values())
    bytes_per_rank = total_size // size
    group, gsize = max(group_sizes.items(), key=lambda x: x[1])
    variables = grouped_data_vars.pop(group)

    assert all(
        variables[0].shape[-1] == v.shape[-1] for v in variables
    ), "All variables in a group need to have the same spatial shape"
    assert all(
        variables[0].chunks[-1] == v.chunks[-1] for v in variables
    ), "All variables in a group need to have the same spatial chunk shape"

    if gsize > bytes_per_rank:
        # Compute the number of ranks required for this group
        nranks = gsize // bytes_per_rank

        cell_chunk_size = variables[0].chunks[-1]
        ncells = variables[0].shape[-1]
        nchunks = ceil(ncells / cell_chunk_size)
        if nchunks < nranks:
            logger.warning(
                f"Not enough chunks to distribute work in group {group}."
                " Consider using fewer processes of split up the group."
            )
            nranks = nchunks

        this_group = [[(group, variables, sl)] for sl in _distribute_chunks(variables[0], nranks)]

        return [*this_group, *distribute_work(grouped_data_vars, size - nranks)]
    else:
        del group_sizes[group]
        result = [(group, variables, slice(None))]

        # Add additional groups to this rank until it reaches the byte limit
        while gsize < bytes_per_rank or size == 1:
            if len(group_sizes) == 0:
                break
            group, next_gsize = max(group_sizes.items(), key=lambda x: x[1])
            if gsize + next_gsize > bytes_per_rank and size > 1:
                break
            result.append((group, grouped_data_vars.pop(group), slice(None)))
            del group_sizes[group]
            gsize += next_gsize

        return [result, *distribute_work(grouped_data_vars, size - 1)]
