#!/usr/bin/env python3

import numpy as np
import zarr
from hiopy._distribute_work import distribute_work

try:
    from pprint import pp
except ImportError:
    print("Cannot find pprint falling back to print")
    pp = print


def _create_test_vars():
    store = {}
    time = zarr.array(data=np.arange(10, dtype=int), path="time", store=store)
    time.attrs["axis"] = "T"
    time.attrs["units"] = "seconds since 2000-01-01"

    var1 = zarr.zeros(shape=[10, 10], path="var1", store=store, chunks=(2, 5))
    var1.time = time
    var2 = zarr.zeros(shape=[10, 10], path="var2", store=store, chunks=(2, 5))
    var2.time = time
    for v in (var1, var2):
        v.attrs["_ARRAY_DIMENSIONS"] = ("time", "cell")

    pp(store)
    return var1, var2


def test_sequential():
    print("Test sequential case")
    var1, var2 = _create_test_vars()
    result = distribute_work({"g1": [var1], "g2": [var2]}, 1)
    pp(result)
    assert len(result) == 1
    assert result[0] == [("g1", [var1], slice(None)), ("g2", [var2], slice(None))]


def test_dist_groups():
    print("Test distributing groups")
    var1, var2 = _create_test_vars()
    result = distribute_work({"g1": [var1], "g2": [var2]}, 2)
    pp(result)
    assert len(result) == 2
    assert result[0] == [("g1", [var1], slice(None))]
    assert result[1] == [("g2", [var2], slice(None))]


def test_dist_chunks():
    print("Test distributing chunks")
    var1, var2 = _create_test_vars()
    result = distribute_work({"g": [var1, var2]}, 2)
    pp(result)
    assert len(result) == 2
    assert result[0] == [("g", [var1, var2], slice(0, 5))]
    assert result[1] == [("g", [var1, var2], slice(5, 10))]


if __name__ == "__main__":
    test_sequential()
    test_dist_groups()
    test_dist_chunks()
