#include <cassert>
#include <charconv>
#include <coyote.hpp>
#include <iostream>
#include <memory>
#include <numeric>
#include <semaphore>

#include <mpi.h>
extern "C" {
#include <yac.h>
}

#include "coyoteenv.hpp"

namespace coyote {

    class Coyote::Impl {
      public:
        Impl(const std::string& name) : _name(name) {
            yac_cpredef_comp_instance(
                CoyoteEnv::getInstance().get_yac_instance_id(), name.c_str(), &_yac_comp_id);
        }

        void add_field(
            const std::string& source_comp,
            const std::string& source_grid,
            const std::string& source_field,
            const std::function<void(Event event)>& eventHandler,
            std::optional<std::string> iso_dt,
            std::optional<int> collection_size,
            std::string interpolation_stack_yaml,
            int yac_time_reduction,
            std::vector<double> frac_mask,
            std::optional<std::string> pass_through_name,
            int src_lag,
            std::optional<std::string> weight_filename) {
            auto& cenv = CoyoteEnv::getInstance();
            cenv.ensure_def_comps();

            if (!collection_size) {
                cenv.ensure_sync_def();
                collection_size = yac_cget_field_collection_size_instance(
                    cenv.get_yac_instance_id(),
                    source_comp.c_str(),
                    source_grid.c_str(),
                    source_field.c_str());
            }

            if (!iso_dt) {
                cenv.ensure_sync_def();
                iso_dt = yac_cget_field_timestep_instance(
                    cenv.get_yac_instance_id(),
                    source_comp.c_str(),
                    source_grid.c_str(),
                    source_field.c_str());
            }

            size_t hash = std::hash<std::string>{}(
                iso_dt.value() + std::to_string(yac_time_reduction) + interpolation_stack_yaml);
            char hash_str[2 * sizeof(hash) + 1];
            std::to_chars(hash_str, hash_str + (2 * sizeof(hash)), hash, 16);
            hash_str[2 * sizeof(hash)] = 0;
            std::string tgt_field_name = "target:" + source_field + "(" + hash_str + ")";
            int target_field_id        = -1;
            yac_cdef_field(
                tgt_field_name.c_str(),
                _yac_comp_id,
                &_yac_point_id,
                1,
                collection_size.value(),
                iso_dt.value().c_str(),
                YAC_TIME_UNIT_ISO_FORMAT,
                &target_field_id);
            std::optional<int> source_field_id;
            if (pass_through_name) {
                int sfid = -1;
                yac_cdef_field(
                    pass_through_name.value().c_str(),
                    _yac_comp_id,
                    &_yac_point_id,
                    1,
                    collection_size.value(),
                    iso_dt.value().c_str(),
                    YAC_TIME_UNIT_ISO_FORMAT,
                    &sfid);
                source_field_id = sfid;
            }
            const char* comp_name = yac_cget_component_name_from_field_id(target_field_id);
            const char* grid_name = yac_cget_grid_name_from_field_id(target_field_id);

            int yac_interpolation_stack = -1;
            yac_cget_interp_stack_config_from_string_yaml(
                interpolation_stack_yaml.c_str(), &yac_interpolation_stack);

            int ext_couple_config_id = -1;
            yac_cget_ext_couple_config(&ext_couple_config_id);
            if (weight_filename)
                yac_cset_ext_couple_config_weight_file(
                    ext_couple_config_id, weight_filename.value().c_str());
            yac_cdef_couple_custom_instance(
                cenv.get_yac_instance_id(),
                source_comp.c_str(),
                source_grid.c_str(),
                source_field.c_str(),
                comp_name,
                grid_name,
                tgt_field_name.c_str(),
                iso_dt.value().c_str(),
                YAC_TIME_UNIT_ISO_FORMAT,
                yac_time_reduction,
                yac_interpolation_stack,
                src_lag,
                /*tgt_lag*/ 1,
                ext_couple_config_id);

            yac_cfree_interp_stack_config(yac_interpolation_stack);

            size_t size = collection_size.value() * yac_cget_points_size(_yac_point_id);
            cenv.add_field(std::make_unique<Field>(
                target_field_id,
                source_field_id,
                std::vector<double>(size),
                frac_mask,
                collection_size.value(),
                eventHandler));
        }

        MPI_Comm get_mpi_comm() const {
            MPI_Comm comm;
            yac_cget_comp_comm(_yac_comp_id, &comm);
            return comm;
        }

        void def_grid(std::span<const int> num_vertices_per_cell,
                      std::span<const double> x_vertices,
                      std::span<const double> y_vertices,
                      std::span<const int> cell_to_vertex,
                      std::optional<std::span<const double>> x_cells,
                      std::optional<std::span<const double>> y_cells,
                      std::optional<std::span<const int>> global_cell_idx) {
            assert(x_vertices.size() == y_vertices.size());
            assert(x_cells.has_value() == y_cells.has_value());
            assert(std::accumulate(num_vertices_per_cell.begin(), num_vertices_per_cell.end(), 0) ==
                   cell_to_vertex.size());

            int num_cells = num_vertices_per_cell.size();
            assert(std::accumulate(num_vertices_per_cell.begin(), num_vertices_per_cell.end(), 0) ==
                   cell_to_vertex.size());
            if (x_cells) {
                assert(num_cells == x_cells.value().size());
                assert(num_cells == y_cells.value().size());
            }
            int grid_id = -1;

            // conditionally handle the yac_grid_cloud case based on num_vertices_per_cell == 0 case
            if (num_cells > 0) {
                yac_cdef_grid_unstruct(
                    _name.c_str(),
                    x_vertices.size(),
                    num_cells,
                    const_cast<int*>(num_vertices_per_cell.data()),
                    const_cast<double*>(x_vertices.data()),
                    const_cast<double*>(y_vertices.data()),
                    const_cast<int*>(cell_to_vertex.data()),
                    &grid_id);

                int point_id = -1;
                if (!x_cells) {
                    yac_cdef_points_unstruct(
                        grid_id,
                        x_vertices.size(),
                        YAC_LOCATION_CORNER,
                        const_cast<double*>(x_vertices.data()),
                        const_cast<double*>(y_vertices.data()),
                        &_yac_point_id);
                } else {
                    yac_cdef_points_unstruct(
                        grid_id,
                        num_cells,
                        YAC_LOCATION_CELL,
                        const_cast<double*>(x_cells.value().data()),
                        const_cast<double*>(y_cells.value().data()),
                        &_yac_point_id);
                }
                if (global_cell_idx) {
                    assert(global_cell_idx.value().size() == num_cells &&
                           "size of global_cell_index "
                           "does not match");
                    yac_cset_global_index(
                        global_cell_idx.value().data(), YAC_LOCATION_CELL, grid_id);
                }

            } else {
                // The case where yac is to treat a point cloud as a grid
                // This happens in the cases of measurement campaigns
                if (x_cells.has_value() || global_cell_idx.has_value()) {
                    throw std::runtime_error(std::string("Undefined usage of the function when "
                                                         "number of vertices per cell are 0."));
                }

                yac_cdef_grid_cloud(
                    _name.c_str(),
                    x_vertices.size(),
                    const_cast<double*>(x_vertices.data()),
                    const_cast<double*>(y_vertices.data()),
                    &grid_id);

                // vertices points (needed e.g. for nearest neighbour)
                yac_cdef_points_unstruct(
                    grid_id,
                    x_vertices.size(),
                    YAC_LOCATION_CORNER,
                    const_cast<double*>(x_vertices.data()),
                    const_cast<double*>(y_vertices.data()),
                    &_yac_point_id);
            }
        }

        std::string get_start_datetime() const {
            return yac_cget_start_datetime_instance(CoyoteEnv::getInstance().get_yac_instance_id());
        }

        std::string get_end_datetime() const {
            return yac_cget_end_datetime_instance(CoyoteEnv::getInstance().get_yac_instance_id());
        }

        std::map<std::string, std::map<std::string, std::vector<std::string>>>
        get_all_comp_grid_fields() const {
            auto& cenv = CoyoteEnv::getInstance();
            cenv.ensure_sync_def();

            std::map<std::string, std::map<std::string, std::vector<std::string>>> ret_val;

            const int total_comps =
                yac_cget_nbr_comps_instance(CoyoteEnv::getInstance().get_yac_instance_id());
            const char** available_components =
                (const char**)malloc((size_t)total_comps * sizeof(*available_components));
            yac_cget_comp_names_instance(
                CoyoteEnv::getInstance().get_yac_instance_id(), total_comps, available_components);

            for (int i = 0; i < total_comps; ++i) {
                auto comp = available_components[i];
                if (!comp) {
                    continue;
                }

                const int total_grids = yac_cget_comp_nbr_grids_instance(
                    CoyoteEnv::getInstance().get_yac_instance_id(), comp);
                const char** available_grids =
                    (const char**)malloc((size_t)total_grids * sizeof(*available_grids));
                yac_cget_comp_grid_names_instance(
                    CoyoteEnv::getInstance().get_yac_instance_id(),
                    comp,
                    total_grids,
                    available_grids);

                std::map<std::string, std::vector<std::string>> grid_variables;
                for (int j = 0; j < total_grids; ++j) {
                    auto grid = available_grids[j];
                    if (!grid) {
                        continue;
                    }
                    const int total_fields = yac_cget_nbr_fields_instance(
                        CoyoteEnv::getInstance().get_yac_instance_id(), comp, grid);
                    const char** available_fields =
                        (const char**)malloc((size_t)total_fields * sizeof(*available_fields));
                    yac_cget_field_names_instance(
                        CoyoteEnv::getInstance().get_yac_instance_id(),
                        comp,
                        grid,
                        total_fields,
                        available_fields);
                    std::vector<std::string> available_variables;
                    for (int k = 0; k < total_fields; ++k) {
                        auto field = available_fields[k];
                        if (!field) {
                            continue;
                        }
                        available_variables.push_back(field);
                    }
                    grid_variables.insert(std::make_pair(grid, available_variables));
                }
                ret_val.insert(std::make_pair(comp, grid_variables));
            }
            return ret_val;
        }

      private:
        std::string _name;
        int _yac_comp_id, _yac_point_id;
    };

    Coyote::Coyote(const std::string& name) : pImpl(std::make_unique<Impl>(name)) {}

    Coyote::Coyote(Coyote&&)            = default;
    Coyote::~Coyote()                   = default;
    Coyote& Coyote::operator=(Coyote&&) = default;

    void Coyote::add_field(
        const std::string& source_comp,
        const std::string& source_grid,
        const std::string& source_name,
        const std::function<void(Event)>& eventHandler,
        std::optional<std::string> iso_dt,
        std::optional<int> collection_size,
        std::string interpolation_stack_yaml,
        int yac_time_reduction,
        std::vector<double> frac_mask,
        std::optional<std::string> pass_through_name,
        int src_lag,
        std::optional<std::string> weight_filename) {
        pImpl->add_field(
            source_comp,
            source_grid,
            source_name,
            eventHandler,
            iso_dt,
            collection_size,
            interpolation_stack_yaml,
            yac_time_reduction,
            frac_mask,
            pass_through_name,
            src_lag,
            weight_filename);
    }

    MPI_Comm Coyote::get_mpi_comm() const { return pImpl->get_mpi_comm(); }

    void Coyote::def_grid(
        std::span<const int> num_vertices_per_cell,
        std::span<const double> x_vertices,
        std::span<const double> y_vertices,
        std::span<const int> cell_to_vertex,
        std::optional<std::span<const double>> x_cells,
        std::optional<std::span<const double>> y_cells,
        std::optional<std::span<const int>> global_cell_idx) {
        pImpl->def_grid(
            num_vertices_per_cell,
            x_vertices,
            y_vertices,
            cell_to_vertex,
            x_cells,
            y_cells,
            global_cell_idx);
    }

    std::map<std::string, std::map<std::string, std::vector<std::string>>>
    Coyote::get_all_comp_grid_fields() const {
        return pImpl->get_all_comp_grid_fields();
    }

    std::string Coyote::get_start_datetime() const { return pImpl->get_start_datetime(); }

    std::string Coyote::get_end_datetime() const { return pImpl->get_end_datetime(); }

    void coyote_init(std::string group_name, uint nthreads) {
        CoyoteEnv::_instance = std::make_unique<CoyoteEnv>(group_name, nthreads);
    }

    std::string coyote_start_datetime() { return CoyoteEnv::getInstance().start_datetime(); }
    std::string coyote_end_datetime() { return CoyoteEnv::getInstance().end_datetime(); }

    MPI_Comm coyote_group_comm() { return CoyoteEnv::getInstance().get_group_comm(); }

    void ensure_enddef() { CoyoteEnv::getInstance().ensure_enddef(); }

    int get_field_collection_size(const std::string& source_comp,
                                  const std::string& source_grid,
                                  const std::string& field_name) {
        return yac_cget_field_collection_size_instance(
            CoyoteEnv::getInstance().get_yac_instance_id(),
            source_comp.c_str(),
            source_grid.c_str(),
            field_name.c_str());
    }

    std::string get_field_timestep(const std::string& source_comp,
                                   const std::string& source_grid,
                                   const std::string& field_name) {
        std::string ret_val  = "";
        const char* timestep = yac_cget_field_timestep_instance(
            CoyoteEnv::getInstance().get_yac_instance_id(),
            source_comp.c_str(),
            source_grid.c_str(),
            field_name.c_str());

        if (timestep) {
            ret_val = std::string(timestep);
        }

        return ret_val;
    }

    std::string get_field_metadata(const std::string& source_comp,
                                   const std::string& source_grid,
                                   const std::string& field_name) {
        const char* metadata = yac_cget_field_metadata_instance(
            CoyoteEnv::getInstance().get_yac_instance_id(),
            source_comp.c_str(),
            source_grid.c_str(),
            field_name.c_str());
        if (metadata) {
            return std::string(metadata);
        }
        return "";
    }

    void coyote_run() { CoyoteEnv::getInstance().run(); }

    void coyote_add_task(std::function<void()> task, int prio) {
        CoyoteEnv::getInstance().add_task(task, prio);
    }

} // namespace coyote
